@@ -1,11 +1,11 @@
-import functools
+import collections
 import re
-from abc import ABCMeta, abstractmethod
+from abc import ABC, abstractmethod
 
 import flask_jwt_extended
 from apispec import APISpec
 from apispec.ext.marshmallow import MarshmallowPlugin
-from flask import jsonify, request
+from flask import jsonify, request, Blueprint
 from flask.helpers import make_response, url_for
 from flask.views import MethodView
 from flask_jwt_extended import (JWTManager, create_access_token,
@@ -18,12 +18,14 @@ from webargs.flaskparser import parser, use_kwargs
 __all__ = [
     'Resource',
     'CreateResource',
+    'ListResource',
     'NonListableRetrieveResource',
     'RetrieveResource',
     'UpdateResource',
     'DeleteResource',
     'CrudResource',
     'RestApi',
+    'RestApiBlueprint',
     'Response',
     'HeadedResponse',
     'WrappedResponse',
@@ -42,7 +44,7 @@ __all__ = [
 ]
 
 
-class Resource(metaclass=ABCMeta):
+class Resource(ABC):
     pass
 
 
@@ -53,6 +55,13 @@ class CreateResource(Resource):
         raise NotImplementedError()
 
 
+class ListResource(Resource):
+
+    @abstractmethod
+    def list(self, *args, **kwargs):
+        raise NotImplementedError()
+
+
 class NonListableRetrieveResource(Resource):
 
     @abstractmethod
@@ -60,14 +69,21 @@ class NonListableRetrieveResource(Resource):
         raise NotImplementedError()
 
 
-class RetrieveResource(NonListableRetrieveResource):
+class RetrieveResource(NonListableRetrieveResource, ListResource, ABC):
+    pass
+
+
+class ReplaceResource(Resource):
 
     @abstractmethod
-    def list(self, *args, **kwargs):
+    def replace(self, *args, **kwargs):
         raise NotImplementedError()
 
 
-class UpdateResource(Resource):
+class UpdateResource(ReplaceResource):
+
+    def replace(self, *args, **kwargs):
+        return self.update(*args, **kwargs)
 
     @abstractmethod
     def update(self, *args, **kwargs):
@@ -81,7 +97,7 @@ class DeleteResource(Resource):
         raise NotImplementedError()
 
 
-class CrudResource(CreateResource, RetrieveResource, UpdateResource, DeleteResource):
+class CrudResource(CreateResource, RetrieveResource, UpdateResource, DeleteResource, ABC):
     pass
 
 
@@ -125,6 +141,16 @@ class HeadedResponse(Response):
     any other given headers to be added as well.
     """
 
+    CORS_SAFELISTED_HEADERS = {
+        'Cache-Control',
+        'Content-Language',
+        'Content-Length',
+        'Content-Type',
+        'Expires',
+        'Last-Modified',
+        'Pragma'
+    }
+
     def __init__(self, data, status_code=None, links=None, headers=None):
         super().__init__(data, status_code, links)
         self._headers = headers
@@ -135,12 +161,17 @@ class HeadedResponse(Response):
             f'<{u}>; rel="{n}"'
             for (n, u) in self.extend_links(base_links).items()
         )
+        added_headers = set()
         if link_header:
             response.headers['Link'] = link_header
-            response.headers['Access-Control-Expose-Headers'] = 'Link'
+            added_headers.add('Link')
         if self._headers is not None:
             for key, value in self._headers.items():
                 response.headers[key] = value
+                added_headers.add(key)
+        expose_headers = added_headers - self.CORS_SAFELISTED_HEADERS
+        if expose_headers:
+            response.headers['Access-Control-Expose-Headers'] = ', '.join(expose_headers)
         return response
 
 
@@ -151,15 +182,18 @@ class WrappedResponse(Response):
     different top-level members.
     """
 
-    def __init__(self, data, status_code=None, links=None, **kwargs):
+    def __init__(self, data, status_code=None, links=None, data_key='data', **kwargs):
         super().__init__(data, status_code, links)
+        self._data_key = data_key
         self._kwargs = kwargs
     
     def generate(self, schema_cls, many, base_links=None):
         return jsonify(
-            data=self.dump_data(schema_cls, many=many),
             links=self.extend_links(base_links),
-            **self._kwargs
+            **{
+                self._data_key: self.dump_data(schema_cls, many=many),
+                **self._kwargs
+            }
         )
 
 
@@ -167,7 +201,7 @@ class RestView(MethodView):
 
     @staticmethod
     def _extract_parent_ids(resource, kwargs):
-        return {p: kwargs[p] for p in resource.id_params[:-1]}
+        return {p.name: kwargs[p.name] for p in resource.id_params[:-1]}
 
     def __init__(self, schema_cls, resource, num_ids=1):
         super().__init__()
@@ -186,7 +220,7 @@ class RestView(MethodView):
         if not isinstance(response, Response):
             response = Response(data=response)
         return response.generate(self.schema_cls, many=False, base_links={
-            'collection': url_for(self.resource.name, _external=True, **parent_ids)
+            'collection': url_for('.' + self.resource.name, _external=True, **parent_ids)
         }), response.get_status_code(default=201)
 
     def get(self, **kwargs):
@@ -196,7 +230,7 @@ class RestView(MethodView):
             if not isinstance(response, Response):
                 response = Response(data=response)
             return response.generate(self.schema_cls, many=True, base_links={
-                'self': url_for(self.resource.name, _external=True, **parent_ids)
+                'self': url_for('.' + self.resource.name, _external=True, **parent_ids)
             }), response.get_status_code(default=200)
         else:
             response = self.resource.retrieve(**kwargs)
@@ -204,18 +238,32 @@ class RestView(MethodView):
                 response = Response(data=response)
             return response.generate(self.schema_cls, many=False, base_links={
                 'self': url_for(
-                    self.resource.name,
+                    '.' + self.resource.name,
                     _external=True,
                     **{**parent_ids, **kwargs}
                 ),
                 'collection': url_for(
-                    self.resource.name,
+                    '.' + self.resource.name,
                     _external=True,
                     **parent_ids
                 )
             }), response.get_status_code(default=200)
 
     def put(self, **kwargs):
+        parent_ids = self._extract_parent_ids(self.resource, kwargs)
+        kwargs.update(parser.parse(
+            self.schema_cls(),
+            request,
+            location='json_or_form'
+        ))
+        response = self.resource.replace(**kwargs)
+        if not isinstance(response, Response):
+            response = Response(data=response)
+        return response.generate(self.schema_cls, many=False, base_links={
+            'collection': url_for('.' + self.resource.name, _external=True, **parent_ids)
+        }), response.get_status_code(default=200)
+
+    def patch(self, **kwargs):
         parent_ids = self._extract_parent_ids(self.resource, kwargs)
         kwargs.update(parser.parse(
             self.schema_cls(partial=True),
@@ -226,7 +274,7 @@ class RestView(MethodView):
         if not isinstance(response, Response):
             response = Response(data=response)
         return response.generate(self.schema_cls, many=False, base_links={
-            'collection': url_for(self.resource.name, _external=True, **parent_ids)
+            'collection': url_for('.' + self.resource.name, _external=True, **parent_ids)
         }), response.get_status_code(default=200)
 
     def delete(self, **kwargs):
@@ -236,11 +284,23 @@ class RestView(MethodView):
 
 class RestApi:
 
-    RE_URL = re.compile(r'<(?:[^:<>]+:)?([^<>]+)>')
+    URL_CONVERTER_TO_TYPE = {
+        'string': 'string',
+        'int': 'integer',
+        'float': 'number',
+        'path': 'string',
+    }
 
-    def __init__(self, app, title, version='v1', spec_path='/spec', docs_path='/docs'):
+    RE_URL = re.compile(r'<(?:([^:<>]+):)?([^<>]+)>')
+
+    IdParam = collections.namedtuple('IdParam', ['type', 'name'])
+
+    def __init__(
+        self, app, title, version='v1', spec_path='/spec', docs_path='/docs', servers=None,
+        security_schemes=None, default_security_scheme=None
+    ):
         self.app = app
-        self.resource_methods = {}
+        self.resource_methods = collections.defaultdict(set)
 
         marshmallow_plugin = MarshmallowPlugin()
         self.spec = spec = APISpec(
@@ -251,7 +311,7 @@ class RestApi:
         )
         self.openapi = marshmallow_plugin.converter  # openapi converter from marshmallow plugin
 
-        self.jwt = JWTManager(app)
+        self.jwt = JWTManager()
         self.token_in_blacklist_checker = self.jwt.token_in_blacklist_loader  # shortcut to blacklisting decorator
         self.token_claims_loader = self.jwt.user_claims_loader  # shortcut to user claims decorator
 
@@ -270,11 +330,28 @@ class RestApi:
             'bearerFormat': 'JWT'
         })
 
+        if security_schemes is not None:
+            for name, scheme in security_schemes.items():
+                spec.components.security_scheme(name, scheme)
+
+        self.default_security_scheme = default_security_scheme
+
+        if not isinstance(app, Blueprint):
+            self.init_app(app, spec_path, docs_path, servers=servers)  # needs to be called separately when blueprint
+
+    def init_app(self, app, spec_path='/spec', docs_path='/docs', docs_blueprint_name='swagger_ui', servers=None):
+        """
+        This needs to be called with the main app as argument when the api is defined on a flask blueprint.
+        """
         @app.route(spec_path)
         def get_spec():
-            return jsonify(spec.to_dict())
+            spec_dict = self.spec.to_dict()
+            return jsonify({'servers': servers, **spec_dict} if servers is not None else spec_dict)
 
-        swaggerui_blueprint = get_swaggerui_blueprint(docs_path, spec_path)
+        self.jwt.init_app(app)
+        swaggerui_blueprint = get_swaggerui_blueprint(
+            docs_path, spec_path, blueprint_name=docs_blueprint_name
+        )
         app.register_blueprint(swaggerui_blueprint, url_prefix=docs_path)
 
     def resource(self, path, name, schema):
@@ -287,18 +364,23 @@ class RestApi:
 
     def add_resource(self, cls, path, name, schema):
         cls.name = name
-        cls.id_params = self.RE_URL.findall(path)
+        cls.id_params = [
+            self.IdParam(self.URL_CONVERTER_TO_TYPE.get(type_, 'string'), name)
+            for (type_, name) in self.RE_URL.findall(path)
+        ]
 
         base_path = '/'.join(path.split('/')[:-1])
         view = RestView.as_view(name, schema, cls(), len(cls.id_params))
         if issubclass(cls, CreateResource):
+            self.resource_methods[name].add('POST')
             self.add_path(base_path, view, method='POST',
                             tag=name, id_params=cls.id_params[:-1],
-                            input_schema=schema, output_schema=schema,
+                            input_schema=schema(), output_schema=schema(),
                             extra_args=getattr(cls.create, '__extra_args__', None),
                             auth_required=getattr(cls.create, '__auth_required__', None),
                             status_code=201, description=cls.create.__doc__)
-        if issubclass(cls, RetrieveResource):
+        if issubclass(cls, ListResource):
+            self.resource_methods[name].add('GET')
             self.add_path(base_path, view, method='GET',
                             tag=name, id_params=cls.id_params[:-1],
                             output_schema=schema(many=True),
@@ -306,47 +388,58 @@ class RestApi:
                             auth_required=getattr(cls.list, '__auth_required__', None),
                             description=cls.list.__doc__)
         if issubclass(cls, NonListableRetrieveResource):
+            self.resource_methods[name].add('GET')
             self.add_path(path, view, method='GET',
                             tag=name, id_params=cls.id_params,
-                            output_schema=schema,
+                            output_schema=schema(),
                             extra_args=getattr(cls.retrieve, '__extra_args__', None),
                             auth_required=getattr(cls.retrieve, '__auth_required__', None),
                             description=cls.retrieve.__doc__)
-        if issubclass(cls, UpdateResource):
+        if issubclass(cls, ReplaceResource):
+            self.resource_methods[name].add('PUT')
             self.add_path(path, view, method='PUT',
                             tag=name, id_params=cls.id_params,
-                            input_schema=schema, output_schema=schema,
+                            input_schema=schema(), output_schema=schema(),
                             extra_args=getattr(cls.update, '__extra_args__', None),
                             auth_required=getattr(cls.update, '__auth_required__', None),
                             description=cls.update.__doc__)
             self.app.add_url_rule(path, view_func=view, methods=['PUT'])
+        if issubclass(cls, UpdateResource):
+            self.resource_methods[name].add('PATCH')
+            self.add_path(path, view, method='PATCH',
+                            tag=name, id_params=cls.id_params,
+                            input_schema=schema(partial=True), output_schema=schema(),
+                            extra_args=getattr(cls.update, '__extra_args__', None),
+                            auth_required=getattr(cls.update, '__auth_required__', None),
+                            description=cls.update.__doc__)
+            self.app.add_url_rule(path, view_func=view, methods=['PATCH'])
         if issubclass(cls, DeleteResource):
+            self.resource_methods[name].add('DELETE')
             self.add_path(path, view, method='DELETE',
                             tag=name, id_params=cls.id_params,
                             extra_args=getattr(cls.delete, '__extra_args__', None),
                             auth_required=getattr(cls.delete, '__auth_required__', None),
                             status_code=204, description=cls.delete.__doc__)
-
-        # keep track of methods
-        self.resource_methods[name] = next(
-            r.methods for r in self.app.url_map.iter_rules() if r.endpoint == name
-        )
            
         return cls
 
     def add_path(self, path, view, method, tag, id_params=None,
                  input_schema=None, output_schema=None, extra_args=None, auth_required=None,
                  status_code=200, description=''):
-        swagger_path = self.RE_URL.sub(r'{\1}', path)
+        swagger_path = self.RE_URL.sub(r'{\2}', path)
         self.app.add_url_rule(path, view_func=view, methods=[method])
 
-        parameters = [{'name': id_param, 'in': 'path'} for id_param in id_params]
+        parameters = [
+            {'name': id_param.name, 'schema': {'type': id_param.type}, 'in': 'path'}
+            for id_param in id_params
+        ]
         if extra_args:
             parameters.extend(self.openapi.fields2parameters(extra_args, default_in='query'))
 
         request_body = {}
         if input_schema:
             request_body['requestBody'] = {
+                'description': '',
                 'content': {
                     'application/json': {
                         'schema': self.openapi.schema2jsonschema(input_schema)
@@ -354,12 +447,16 @@ class RestApi:
                 }
             }
 
+        if auth_required is None:
+            auth_required = self.default_security_scheme
+
         self.spec.path(swagger_path, operations={
             method.lower(): {
-                'description': description,
+                'description': description or '',
                 'parameters': parameters,
                 'responses': {
                     str(status_code): {} if not output_schema else {
+                        'description': '',
                         'content': {
                             'application/json': {
                                 'schema': output_schema
@@ -385,7 +482,7 @@ class RestApi:
                 if desired_method in methods:
                     _method = desired_method
                     break
-        return url_for(resource_name, _method=_method, _external=_external, **kwargs)
+        return url_for('.' + resource_name, _method=_method, _external=_external, **kwargs)
 
 
 class RestApiBlueprint:
@@ -410,6 +507,10 @@ class RestApiBlueprint:
         self.resources.append((cls, (path, name, schema)))
         return cls
     
+    def add_blueprint(self, blueprint):
+        blueprint.bind(self)
+        self.resources.extend(blueprint.resources)
+
     def url_for(self, resource_name, _method=None, _external=True, **kwargs):
         return self._rest_api.url_for(resource_name, _method=_method, _external=_external, **kwargs)
 
